{
Custom Characters 2.3

(C) 1993 Steve Goldsmith
All Rights Reserved

Palette Adjuster dialog.  See CC.PAS for usage.
}

unit PalDlg;

{$I APP.INC}

interface

uses

  Objects, App, Views, Dialogs, Drivers,
  VGA, Cmds;

type

  PColPalView = ^TColPalView;
  TColPalView = object (TView)
    StartColor : byte;
    constructor Init (var Bounds : TRect; StartCol : byte);
    procedure Draw; virtual;
  end;

  PMsgButtons = ^TMsgButtons;
  TMsgButtons = object (TRadioButtons)
    procedure Press(Item: Integer); virtual;
    procedure MovedTo (Item:Integer); virtual;
  end;

  PPalDlg = ^TPalDlg;
  TPalDlg = object (TDialog)
    CurPal : vgaPalette;
    RedBar,
    GreenBar,
    BlueBar : PScrollBar;
    DefColor : PMsgButtons;
    constructor Init;
    procedure SetColorBars (Color : byte);
    procedure ChangeDAC;
    procedure HandleEvent(var Event: TEvent); virtual;
  end;

implementation

{TColPalView}

constructor TColPalView.Init (var Bounds : TRect; StartCol : byte);

begin
  inherited Init (Bounds);
  StartColor := StartCol
end;

procedure TColPalView.Draw;

var

  MiniBuf : word;
  Y, Color : byte;

begin {draw color blocks using '' character}
  Color := GetColor (6) and $f0;
  for Y := 0 to Size.Y-1 do
  begin
    MiniBuf := ((Color or (StartColor+Y)) shl 8) or 219;
    WriteLine (0,Y,1,1,MiniBuf)
  end
end;

{TMsgButtons}

procedure TMsgButtons.Press (Item : integer);

begin
  inherited Press (Item); {send message when radio buttons pressed}
  Message (Owner,evBroadcast,cmMsgButtonPress,nil)
end;

procedure TMsgButtons.MovedTo (Item:Integer);

begin
  inherited MovedTo (Item); {send message when button moved}
  Message (Owner,evBroadcast,cmMsgButtonPress,nil)
end;

{TPalDlg}

constructor TPalDlg.Init;

var

  R : TRect;

begin
  R.Assign (0,0,40,20);
  inherited Init (R,'Adjust Palette');
  R.Assign(2, 2, 38, 10);
  DefColor := New(PMsgButtons, Init(R, {use bios default color names}
    NewSItem('Black',
    NewSItem('Blue',
    NewSItem('Green',
    NewSItem('Cyan',
    NewSItem('Red',
    NewSItem('Magenta',
    NewSItem('Brown',
    NewSItem('Light Gray',
    NewSItem('Gray',
    NewSItem('Light Blue',
    NewSItem('Light Green',
    NewSItem('Light Cyan',
    NewSItem('Light Red',
    NewSItem('Light Magenta',
    NewSItem('Yellow',
    NewSItem('White',
    nil))))))))))))))))));
  Insert(DefColor);
  R.Assign(1,1,7,2);
  Insert(New(PLabel, Init(R, '~C~olors', DefColor)));
  R.Assign(6, 2, 7, 10);
  Insert(New(PColPalView, Init(R,0)));
  R.Assign(22, 2, 23, 10);
  Insert(New(PColPalView, Init(R,8)));
  R.Assign(2,11,36,12);
  RedBar := New (PScrollBar, Init (R));
  RedBar^.SetParams (0,0,63,8,1);
  RedBar^.Options := RedBar^.Options or ofSelectable;
  Insert (RedBar);
  R.Assign(1,10,5,11);
  Insert (New (PLabel,Init (R,'~R~ed',RedBar)));
  R.Assign(2,13,36,14);
  GreenBar := New (PScrollBar, Init (R));
  GreenBar^.SetParams (0,0,63,8,1);
  GreenBar^.Options := GreenBar^.Options or ofSelectable;
  Insert (GreenBar);
  R.Assign(1,12,7,13);
  Insert (New (PLabel,Init (R,'~G~reen',GreenBar)));
  R.Assign(2,15,36,16);
  BlueBar := New (PScrollBar, Init (R));
  BlueBar^.SetParams (0,0,63,8,1);
  BlueBar^.Options := BlueBar^.Options or ofSelectable;
  Insert (BlueBar);
  R.Assign(1,14,6,15);
  Insert (New (PLabel,Init (R,'~B~lue',BlueBar)));
  R.Assign(1, 17, 11, 19);
  Insert(New(PButton, Init(R, 'O~K~', cmOk, bfDefault)));
  R.Assign(12, 17, 22, 19);
  Insert(New(PButton, Init(R, 'Cancel', cmCancel, bfNormal)));
  R.Assign(23, 17, 36, 19);
  Insert(New(PButton, Init(R, '~D~efault', cmDefaultPal, bfNormal)));
  GetDACBlock (@CurPal,0,256);
  SetColorBars (0);
  SelectNext (false)
end;

procedure TPalDlg.SetColorBars (Color : byte);

var

  R,G,B : byte;

begin
  GetDAC (GetAttrCont (Color),R,G,B); {set scroll bars to rgb}
  RedBar^.SetValue (R);
  GreenBar^.SetValue (G);
  BlueBar^.SetValue (B)
end;

procedure TPalDlg.ChangeDAC;

var

  R,G,B : byte;

begin {find which scroll bar is selected and set that dac register}
  GetDAC (GetAttrCont (DefColor^.Value),R,G,B);
  if RedBar^.State and sfSelected = sfSelected then
    R := RedBar^.Value
  else
    if GreenBar^.State and sfSelected = sfSelected then
      G := GreenBar^.Value
    else
      if BlueBar^.State and sfSelected = sfSelected then
        B := BlueBar^.Value;
  SetDAC (GetAttrCont (DefColor^.Value),R,G,B)
end;

procedure TPalDlg.HandleEvent(var Event: TEvent);

begin
  if (Event.What = evCommand) and
  ((Event.Command = cmCancel) or
  (Event.Command = cmClose)) then
  SetDACBlock (@CurPal,0,256); {set default palette}
  inherited HandleEvent(Event);
  case Event.What of
    evCommand:
      begin
        case Event.Command of
          cmOk         : Close;
          cmDefaultPal :
          begin
            SetDACBlock (@CurPal,0,256); {set default palette}
            SetColorBars (DefColor^.Value) {set scroll bars too}
          end
        else
          Exit
        end;
        ClearEvent (Event)
      end;
    evBroadcast:
      begin {process broadcasts}
        case Event.Command of
          cmScrollBarChanged : ChangeDAC;
          cmMsgButtonPress : SetColorBars (DefColor^.Value)
        else
          Exit
        end;
        ClearEvent (Event)
      end
  end
end;

end.
