{
Custom Characters 2.3

(C) 1993 Steve Goldsmith
All Rights Reserved

Fast PCX objects for encoding and decoding various PCX formats.

Currently supports:

Decode/Encode 2 color.
X size <= 640
Y size <= 480

Decode/Encode 256 color/gray.
X size <= 320
Y size <= 200.
}

unit PCX;

{$I APP.INC}
{$X+}

interface

uses

  Dos, Objects, VGA;

const

  pcxManufacturer = $0a; {used to identify .pcx}
  pcxStart256Pal = $0c;  {byte before 256 color palette}

{pcx Versions}

  pcxVer25 = 0; pcxVer28Pal = 2; pcxVer28NoPal = 3; pcxVer30 = 5;

{pcx palette types}

  pcxColorPal = 1; pcxGrayPal = 2;

{pcx max encode/decode sizes}

  pcxMaxXSize1 = 640;
  pcxMaxYSize1 = 480;
  pcxMaxXSize256 = 320;
  pcxMaxYSize256 = 200;
  pcxMaxEncodeSize = 4096; {encode buffer size}
  pcxMaxDecodeSize = 4096; {decode buffer size}

{pcx error codes}

  pcxMemAlloc      = -100;
  pcxNotPCXformat  = -101;
  pcxNot2Color     = -102;
  pcxNot256Color   = -103;
  pcxXSize         = -104;
  pcxYSize         = -105;

type

{128 byte pcx header record}

  pcx16Pal = array[0..47] of byte;
  pcxHeader = record
    Manufacturer,
    Version,
    Encoding,
    BitsPerPixel : byte;
    XMin,
    YMin,
    XMax,
    YMax,
    HRes,
    VRes : integer;
    Palette : pcx16Pal;
    Reserved,
    ColorPlanes : byte;
    BytesPerLine,
    PaletteType : integer;
    Filler : array[0..57] of byte;
  end;

  PReadPCXFile = ^TReadPCXFile;

  TReadPCXFile = object (TObject)
    ReadError : integer;
    XSize,
    YSize,
    EncodeBufPos : word;
    ReadFileSize,
    EncodeSize : longint;
    ReadFile : file;
    EncodeBufPtr : vgaDataBufPtr;
    DecodeBufPtr : vgaDataBufPtr;
    Header : pcxHeader;
    ReadPalette : vgaPalette;
    constructor Init (FileName : PathStr);
    destructor Done; virtual;
    function GetEncodeByte : byte;
    procedure DecodeFile; virtual;
  end;

  PDecodePCXFile2 = ^TDecodePCXFile2;

  TDecodePCXFile2 = object (TReadPCXFile)
    constructor Init (FileName : PathStr);
  end;

  PDecodePCXFile256 = ^TDecodePCXFile256;

  TDecodePCXFile256 = object (TReadPCXFile)
    constructor Init (FileName : PathStr);
    procedure ReadPal256;
    procedure DecodeFile; virtual;
    procedure Palette256to64;
  end;

  PWritePCXFile = ^TWritePCXFile;

  TWritePCXFile = object (TObject)
    WriteError : integer;
    EncodeBufPos : word;
    WriteFile : file;
    EncodeBufPtr : vgaDataBufPtr;
    Header : pcxHeader;
    constructor Init (FileName : PathStr);
    destructor Done; virtual;
    procedure WriteHeader;
    procedure SetHeader (XS, YS : integer; BP, CP : byte);
    procedure WriteEncodeByte (E : byte);
    procedure FlushEncodeBuf;
    procedure EncodeFile (PCXImage : vgaDataBufPtr); virtual;
  end;

  PEncodePCXFile2 = ^TEncodePCXFile2;

  TEncodePCXFile2 = object (TWritePCXFile)
    procedure EncodeFile (PCXImage : vgaDataBufPtr); virtual;
  end;

  PEncodePCXFile256 = ^TEncodePCXFile256;

  TEncodePCXFile256 = object (TWritePCXFile)
    procedure EncodeFile (PCXImage : vgaDataBufPtr); virtual;
    procedure WritePal256 (Pal : vgaPalettePtr);
    procedure Palette64to256 (Pal : vgaPalettePtr);
  end;

const
  {default rgb triples for 1 bit images}
  pcxDef1BitPal : pcx16Pal = (
  $00,$00,$00,
  $ff,$ff,$ff,
  $00,$aa,$00,
  $00,$aa,$aa,
  $aa,$00,$00,
  $aa,$00,$aa,
  $aa,$aa,$00,
  $aa,$aa,$aa,
  $55,$55,$55,
  $55,$55,$ff,
  $55,$ff,$55,
  $55,$ff,$ff,
  $ff,$55,$55,
  $ff,$55,$ff,
  $ff,$ff,$55,
  $ff,$ff,$ff
  );

implementation

end.
